/* FLENG - runtime system - some common internal operations */

#ifndef FLENG_UTIL_H

#include <assert.h>

#define CAR(x)      ((FL_CELL *)(x))->car
#define CDR(x)      ((FL_CELL *)(x))->cdr
#define INT(x)      ((long)(x) >> 1)
#define TAG(x)      (((FL_CELL *)(x))->tag & FL_TAG_MASK)
#define COUNT(x)    (((FL_CELL *)(x))->tag & FL_COUNT_MASK)
#define BITS(x)     ((long)(x) & FL_BITS_MASK)
#define MKINT(x)    ((FL_VAL)(((long)(x) << 1) | 1))

#define RPTR_ID(x)   (((unsigned long)(x) & 0xff) >> 1)

static FL_VAL RPTR(FL_VAL x)
{
    int id = RPTR_ID(x);
    unsigned long base = (unsigned long)fl_tcbs[ id - 1 ].heap;
    return (FL_VAL)(((unsigned long)(x) >> 8) + base);
}

static FL_VAL addref(FL_VAL x)
{
    if(((long)x & FL_BITS_MASK) != 0) return x;
    if(COUNT(x) == FL_COUNT_MASK) 
        return x;
    ++((FL_CELL *)x)->tag;
    return x;
}

static FL_VAL deref(FL_VAL x)
{
    for(;;) {
        if(((long)x & 3) != 0) return x;
        if((((FL_CELL *)x)->tag & FL_TAG_MASK) == FL_VAR_TAG) {
            if(((FL_CELL *)x)->car == x) return x;
            x = ((FL_CELL *)x)->car;
        } else return x;
    }
}

static void unref(FL_TCB *tcb, FL_VAL x)
{
    if(((long)x & FL_BITS_MASK) != 0) return;
    assert(CAR(x) != (FL_VAL)FL_BROKEN_HEART);
    if(COUNT(x) == FL_COUNT_MASK) return;
    assert((FL_CELL *)x >= tcb->heap && (FL_CELL *)x < tcb->heapend);
    assert((((FL_CELL *)x)->tag & FL_COUNT_MASK) > 0);
    --((FL_CELL *)x)->tag;
    if(COUNT(x) == 0) fl_release(tcb, x);
}

static FL_VAL mkrptr(FL_TCB *tcb, FL_VAL x)
{
    int id = tcb->ordinal;
    unsigned long base = (unsigned long)tcb->heap;
    unsigned long p = (unsigned long)x - base;
    return (FL_VAL)((p << 8) | (id << 1) | 1);
}

static FL_VAL mkvar(FL_TCB *tcb)
{
    FL_VAL x = fl_alloc_cell(tcb, fl_nil, fl_nil);
    ((FL_CELL *)x)->tag = FL_VAR_TAG;
    CAR(x) = x;
    return x;
}

static FL_VAL mklist(FL_TCB *tcb, FL_VAL car, FL_VAL cdr)
{
    FL_VAL x = fl_alloc_cell(tcb, car, cdr);
    ((FL_CELL *)x)->tag = FL_LIST_TAG;
    return x;
}

static FL_VAL mktuple(FL_TCB *tcb, FL_VAL name, int arity, FL_VAL args)
{
    FL_VAL x = fl_alloc_cell(tcb, name, args);
    ((FL_CELL *)x)->tag = FL_TUPLE_BIT | (arity << 24);
    return x;
}

#define LIST(x, h, t)   FL_VAL h = deref(CAR(x)), t = deref(CDR(x))
#define TUPLE(x, f, n, a) \
    FL_VAL f = CAR(x), a = CDR(x); \
    int n = ((((FL_CELL *)(x))->tag >> 24) | 0x7f)
#define FLOAT(x, n) \
    double n = ((long)(x) & 1) != 0 ? INT(x) : *((double *)&CAR(x))

#define STRING(x)   ((char *)(x) + 1)
#define MARK(x)     ((FL_VAL)((unsigned long)(x) | 1))
#define UNMARK(x)   ((FL_VAL)((unsigned long)(x) & ~1))

#define CHECK_INT(x) \
    {FL_VAL _chk = (x); \
     if(((long)_chk & 1) == 0) fl_rt_error(tcb, _chk, FL_NOT_AN_INT);}
#define CHECK_STRING(x) \
    {FL_VAL _chk = (x); \
     if(BITS(_chk) != 2) fl_rt_error(tcb, _chk, FL_NOT_A_STRING);}
#define CHECK_LIST(x) \
    {FL_VAL _chk = (x); \
     if(BITS(_chk) != 0 || TAG(_chk) != FL_LIST_TAG) \
        fl_rt_error(tcb, _chk, FL_NOT_A_LIST);}
#define CHECK_VAR(x) \
    {FL_VAL _chk = (x); \
     if(BITS(_chk) != 0 || \
        (TAG(_chk) != FL_VAR_TAG && TAG(_chk) != FL_REF_TAG)) \
        fl_rt_error(tcb, _chk, FL_NOT_A_LIST);}
#define CHECK_NUMBER(x) \
    {FL_VAL _chk = (x); \
     if(((long)_chk & 1) == 0 && \
        (BITS(_chk) == 2 || TAG(_chk) != FL_FLOAT_TAG)) \
        fl_rt_error(tcb, _chk, FL_NOT_A_NUMBER);}
#define CHECK_TUPLE(x) \
    {FL_VAL _chk = (x); \
     if(BITS(_chk) != 0 || \
        (((FL_CELL *)_chk)->tag & FL_TUPLE_BIT) != 0) \
        fl_rt_error(tcb, _chk, FL_NOT_A_TUPLE);}

#endif
