# FLENG - asm part of x86_64 runtime

    .text

    .extern fl_rt_error
    .extern fl_release
    .extern fl_resolve_pdef
    .extern fl_true, fl_false, fl_nil
    .extern fl_log_entry, fl_writeln
    .extern fl_unify_rec
    .extern fl_ordering

    # rdx=error-code, rsi=argument
error: 
    movq %rbx,%rdi
    andq $-16,%rsp
    call fl_rt_error

    .global fl_check_atom
    # rax=val
fl_check_atom: 
    movq %rax,%rdx
    andq $3,%rdx
    cmpq $2,%rdx
    jne fl_check_atom_l1
    ret
fl_check_atom_l1: 
    movq %rax,%rsi
    movq $1,%rdx        # FL_NOT_AN_ATOM
    jmp error

    .global fl_check_cell
    # rax=val
fl_check_cell: 
    testq $3,%rax
    jnz fl_check_cell_l1
    ret
fl_check_cell_l1: 
    movq %rax,%rsi
    movq $2,%rdx        # FL_NOT_A_CELL
    jmp error

    .global fl_check_integer
    # rax=val
fl_check_integer: 
    testq $1,%rax       # FL_INT_BIT
    jz fl_check_integer_l1
    ret
fl_check_integer_l1: 
    movq %rax,%rsi
    movq $3,%rdx        # FL_NOT_AN_INT
    jmp error

    .global fl_check_list
    # rax=val
fl_check_list: 
    testq $3,%rax
    jnz fl_check_list_l1
    movq (%rax),%rdx
    movq $0xff000000,%rcx # FL_TAG_MASK
    andq %rcx,%rdx
    cmpq $0x01000000,%rdx # FL_LIST_TAG
    jne fl_check_list_l1
    ret
fl_check_list_l1: 
    movq %rax,%rsi
    movq $4,%rdx        # FL_NOT_A_LIST
    jmp error

    .global fl_check_module
    # rax=val
fl_check_module: 
    testq $3,%rax
    jnz fl_check_module_l1
    movq (%rax),%rdx
    movq $0xff000000,%rcx # FL_TAG_MASK
    andq %rcx,%rdx
    cmpq $0x08000000,%rdx # FL_MODULE_TAG
    jne fl_check_module_l1
    ret
fl_check_module_l1: 
    movq %rax,%rsi
    movq $23,%rdx        # FL_NOT_A_MODULE
    jmp error

    .global fl_check_tuple
    # rax=val
fl_check_tuple: 
    testq $3,%rax
    jnz fl_check_tuple_l1
    movl (%rax), %r9d
    testl $0x80000000, %r9d
    jnz fl_check_tuple_l1
    ret
fl_check_tuple_l1: 
    movq %rax,%rsi
    movq $5,%rdx        # FL_NOT_A_TUPLE
    jmp error

    .global fl_check_var
    # rax=val
fl_check_var: 
    testq $3,%rax
    jnz fl_check_var_l1
    movl (%rax),%edx
    andl $0xff000000,%edx # FL_TAG_MASK
    cmpl $0x04000000,%edx # FL_VAR_TAG
    je fl_check_var_l1
    cmpl $0x07000000,%edx # FL_REF_TAG
    je fl_check_var_l1
    movq %rax,%rsi
    movq $6,%rdx        # FL_NOT_A_VAR
    jmp error
fl_check_var_l1: 
    ret

    .global fl_check_number
    # rax=val
fl_check_number: 
    testq $1,%rax
    jz fl_check_number_l1
    ret
fl_check_number_l1: 
    testq $2,%rax
    jnz fl_check_number_l2
    movl (%rax),%edx
    andl $0xff000000,%edx # FL_TAG_MASK
    cmpl $0x02000000,%edx # FL_FLOAT_TAG
    jne fl_check_number_l2
    ret
fl_check_number_l2: 
    movq %rax,%rsi
    movq $11,%rdx       # FL_NOT_A_NUMBER
    jmp error

    # rsi=val -> xmm0
float_arg_1: 
    testq $1,%rsi
    jz float_arg_1_l1
    sarq %rsi
    cvtsi2sd %rsi, %xmm0
    ret
float_arg_1_l1: 
    movsd 8(%rsi), %xmm0
    ret

    # rdx=val -> xmm1
float_arg_2: 
    testq $1,%rdx
    jz float_arg_2_l1
    sarq %rdx
    cvtsi2sd %rdx, %xmm1
    ret
float_arg_2_l1: 
    movsd 8(%rdx), %xmm1
    ret

    # xmm0=val -> rax (float)
float_result: 
    movq $1,%rax
    movq %rax, %r10
    call cons
    movsd %xmm0, 8(%rax)
    movl $0x02000000,(%rax)         # FL_FLOAT_TAG
    ret

    .global fl_add
    # rsi=num, rdx=num -> rax
fl_add: 
    testq $1,%rsi
    jz fl_add_l1
    testq $1,%rdx
    jz fl_add_l1
    leaq -1(%rsi),%rax
    addq %rdx,%rax
    ret
fl_add_l1: 
    call float_arg_1
    call float_arg_2
    addsd %xmm1, %xmm0
    jmp float_result

    .global fl_sub
    # rsi=num, rdx=num -> rax
fl_sub: 
    testq $1,%rsi
    jz fl_sub_l1
    testq $1,%rdx
    jz fl_sub_l1
    subq %rdx,%rsi
    leaq 1(%rsi),%rax
    ret
fl_sub_l1: 
    call float_arg_1
    call float_arg_2
    subsd %xmm1, %xmm0
    jmp float_result

    .global fl_mul
    # rsi=num, rdx=num -> rax
fl_mul: 
    testq $1,%rsi
    jz fl_mul_l1
    testq $1,%rdx
    jz fl_mul_l1
    sarq %rsi
    sarq %rdx
    imulq %rdx,%rsi
    movq %rsi,%rax
    shlq %rax
    orq $1,%rax
    ret
fl_mul_l1: 
    call float_arg_1
    call float_arg_2
    mulsd %xmm1, %xmm0
    jmp float_result

    .global fl_div
    # rsi=num, rdx=num -> rax
fl_div: 
    cmpq $1,%rdx
    jne fl_div_l0
    movq $12,%rsi       # FL_DIV_BY_ZERO
    jmp error
fl_div_l0: 
    testq $1,%rsi
    jz fl_div_l1
    testq $1,%rdx
    jz fl_div_l1
    sarq %rsi
    sarq %rdx
    movq %rsi,%rax
    movq %rdx,%rcx
    cqto
    idivq %rcx
    shlq %rax
    orq $1,%rax
    ret
fl_div_l1: 
    call float_arg_1
    call float_arg_2
    divsd %xmm1, %xmm0
    jmp float_result

    .global fl_mod
    # rsi=num, rdx=num -> rax
fl_mod: 
    cmpq $1,%rdx
    jne fl_mod_l0
    movq $12,%rsi       # FL_DIV_BY_ZERO
    jmp error
fl_mod_l0: 
    testq $1,%rsi
    jz fl_mod_l1
    testq $1,%rdx
    jz fl_mod_l1
    sarq %rsi
    sarq %rdx
    movq %rsi,%rax
    movq %rdx,%rcx
    cqto
    idivq %rcx
    shlq %rdx
    orq $1,%rdx
    movq %rdx,%rax
    ret
fl_mod_l1: 
    call float_arg_1
    call float_arg_2
    divsd %xmm1, %xmm0
    jmp float_result

    .global fl_sqrt
    # rsi=num -> rax
fl_sqrt: 
    call float_arg_1
    movsd %xmm0, -8(%rsp)
    fldl -8(%rsp)
    fsqrt
    fstpl -8(%rsp)
    movsd -8(%rsp), %xmm0
    jmp float_result

    .global fl_sin
    # rsi=num -> rax
fl_sin: 
    call float_arg_1
    .extern sin
    pushq %rax  # dummy
    call sin
    addq $8,%rsp
    jmp float_result

    .global fl_cos
    # rsi=num -> rax
fl_cos: 
    call float_arg_1
    .extern cos
    pushq %rax  # dummy
    call cos
    addq $8,%rsp
    jmp float_result

    .global fl_truncate
    # rsi=num -> rax
fl_truncate: 
    call float_arg_1
    .extern trunc
    pushq %rax  # dummy
    call trunc
    addq $8,%rsp
    jmp float_result

    .global fl_floor
    # rsi=num -> rax
fl_floor: 
    call float_arg_1
    .extern floor
    pushq %rax  # dummy
    call floor
    addq $8,%rsp
    jmp float_result

    .global fl_round
    # rsi=num -> rax
fl_round: 
    call float_arg_1
    .extern round
    pushq %rax  # dummy
    call round
    addq $8,%rsp
    jmp float_result

    .global fl_ceiling
    # rsi=num -> rax
fl_ceiling: 
    call float_arg_1
    .extern ceil
    pushq %rax  # dummy
    call ceil
    addq $8,%rsp
    jmp float_result

    .global fl_shl
    # rsi=int, rdx=int -> rax(int)
fl_shl: 
    movq %rsi,%rax
    movq %rdx,%rcx
    shr %rcx
    andq $~1, %rax
    shl %cl,%rax
    orq $1,%rax
    ret

    .global fl_shr
    # rsi=int, rdx=int -> rax(int)
fl_shr: 
    movq %rsi,%rax
    movq %rdx,%rcx
    shr %rcx
    sar %cl,%rax
    orq $1,%rax
    ret

    .global fl_int
    # rsi=num -> rax
fl_int: 
    testq $1,%rsi
    jz fl_int_l1
    movq %rsi,%rax
    ret
fl_int_l1: 
    fldl 8(%rsi)
    fisttpq -8(%rsp)
    movq -8(%rsp),%rax
    shlq %rax
    orq $1,%rax
    ret

    .global fl_float
    # rsi=num -> rax
fl_float: 
    call float_arg_1
    call float_result
    ret

true: 
    movq fl_true,%rax
    ret

false: 
    movq fl_false,%rax
    ret

    .global fl_sametype
    # rsi=val, rdx=val -> rax
fl_sametype: 
    movq %rsi,%rax
    andq $3,%rax
    movq %rdx, %r9
    andq $3, %r9
    cmpq %r9, %rax
    jne false
    testq $3,%rax
    jnz true
    movl (%rsi),%eax
    movl (%rdx), %r9d
    testl $0x80000000,%eax  # FL_TUPLE_BIT
    jz fl_sametype_l4
    testl $0x80000000, %r9d
    jnz false
    jmp true
fl_sametype_l4: 
    testl $0x80000000, %r9d
    jnz false
    andl $0xff000000,%eax   # FL_TAG_MASK
    andl $0xff000000, %r9d
    cmpl %r9d, %eax
    jne false
    jmp true

    .global fl_equal
    # rsi=val, rdx=val -> rax
    # must preserve P0-P3
fl_equal: 
    cmpq %rdx,%rsi
    je true
    testq $1,%rsi
    jz fl_equal_l2
    testq $1,%rdx
    jz fl_equal_l3
    jmp false
fl_equal_l2: 
    testq $2,%rsi
    jnz false
    movl (%rsi), %r9d
    andl $0xff000000, %r9d
    cmpl $0x02000000, %r9d
    jne false
fl_equal_l3: 
    testq $2,%rdx
    jnz false
    testq $1,%rdx
    jnz fl_equal_l5
    movl (%rdx), %r9d
    andl $0xff000000, %r9d
    cmpl $0x02000000, %r9d
    jne false
fl_equal_l5: 
    call float_arg_1
    call float_arg_2
    ucomisd %xmm1, %xmm0
    jne false
    jmp true

    .global fl_less
    # rsi=val, rdx=val -> rax
    # must preserve P0-P3
fl_less: 
    testq $1,%rsi
    jz fl_less_l1
    testq $1,%rdx
    jz fl_less_l1
    movq fl_true,%rax
    cmpq %rdx,%rsi
    cmovgeq fl_false, %rax
    ret
fl_less_l1: 
    call float_arg_1
    call float_arg_2
    movq fl_true,%rax
    ucomisd %xmm1, %xmm0
    cmovaeq fl_false, %rax
    ret

    .global fl_below
    # rsi=val, rdx=val -> rax
    # must preserve P0-P3
fl_below: 
    pushq %rsi
    pushq %rdx
    pushq %rcx
    pushq %r8
    movq %rsi,%rdi
    movq %rdx,%rsi
    call fl_ordering
    popq %r8
    popq %rcx
    popq %rdx
    popq %rsi
    testq $2,%rax   # -1
    jnz true
    jmp false

    .global fl_above
    # rsi=val, rdx=val -> rax
    # must preserve P0-P3
fl_above: 
    pushq %rsi
    pushq %rdx
    pushq %rcx
    pushq %r8
    movq %rsi,%rdi
    movq %rdx,%rsi
    call fl_ordering
    popq %r8
    popq %rcx
    popq %rdx
    popq %rsi
    cmpq $1,%rax
    jne false
    jmp true

    .global fl_max
    # rsi=val, rdx=val -> rax
fl_max: 
    # must preserve P0-P3
    testq $1,%rsi
    jz fl_max_l1
    testq $1,%rdx
    jz fl_max_l1
    movq %rsi,%rax
    cmpq %rdx,%rsi
    cmovleq %rdx, %rax
    ret
fl_max_l1: 
    pushq %rsi
    pushq %rdx
    call float_arg_1
    call float_arg_2
    popq %rdx
    popq %rax
    ucomisd %xmm1, %xmm0
    cmovbeq %rdx, %rax
    ret

    .global fl_min
    # rsi=val, rdx=val -> rax
fl_min: 
    # must preserve P0-P3
    testq $1,%rsi
    jz fl_min_l1
    testq $1,%rdx
    jz fl_min_l1
    movq %rsi,%rax
    cmpq %rdx,%rsi
    cmovaeq %rdx, %rax
    ret
fl_min_l1: 
    pushq %rsi
    pushq %rdx
    call float_arg_1
    call float_arg_2
    popq %rdx
    popq %rax
    ucomisd %xmm1, %xmm0
    cmovaeq %rdx, %rax
    ret

    .global fl_deref
    # rax=val -> rax
    # must preserve P0-P3
fl_deref: 
    testq $3,%rax
    jz fl_deref_l1
fl_deref_l2: 
    ret
fl_deref_l1: 
    movl (%rax), %r9d
    andl $0xff000000, %r9d
    cmpl $0x04000000, %r9d
    jne fl_deref_l2
    movq 8(%rax), %r9
    cmpq %r9, %rax
    je fl_deref_l2
    movq %r9, %rax
    jmp fl_deref

    .global fl_unref
    # r9=val, rbx=tcb
    # must preserve A + P0-P3
fl_unref: 
    testq $3, %r9
    jz fl_unref_l1
    ret
fl_unref_l1: 
    movl (%r9), %r10d
    andl $0xffffff, %r10d
    cmpl $0xffffff, %r10d
    jne fl_unref_l3
    ret
fl_unref_l3: 
    decl %r10d
    jz fl_unref_l2
    decl (%r9)
    ret
fl_unref_l2: 
    pushq %rax
    pushq %rsi
    pushq %rdx
    pushq %rcx
    pushq %r8
    movq %rbx,%rdi
    movq %r9, %rsi
    call fl_release
    popq %r8
    popq %rcx
    popq %rdx
    popq %rsi
    popq %rax
    ret

setcarry: 
    stc
    ret

clrcarry: 
    clc
    ret

    .global fl_force
    # rax=val -> rax, CF is 1 when suspending
    # rbx=tcb
fl_force: 
    testq $3,%rax
    jnz clrcarry
    movl (%rax),%edx
    andl $0xff000000,%edx # FL_TAG_MASK
    cmpl $0x07000000,%edx # FL_REF_TAG
    je fl_force_l3
    cmpl $0x04000000,%edx # FL_VAR_TAG
    jne clrcarry
    movq 8(%rax),%rdx
    cmpq %rdx,%rax
    je fl_force_l3
    movq %rdx,%rax
    jmp fl_force
fl_force_l3: 
    movq 8(%rbx), %r9
    movq %rax, (%r9)
    movq (%r12), %rax       # goal->addr
    movq %rax, 8(%r9)
    addq $16, %r9
    movq %r9, 8(%rbx)
    jmp setcarry

    .global fl_islist_t
    # rax=val, set CF if not a list
fl_islist_t: 
    testq $3,%rax
    jnz setcarry
    movl (%rax), %r9d
    andl $0xff000000, %r9d
    cmpl $0x01000000, %r9d
    jne setcarry
    jmp clrcarry

    .global fl_istuple_t
    # rax=val, set CF if not a tuple
fl_istuple_t: 
    testq $3,%rax
    jnz setcarry
    movl (%rax),%edx
    testl $0x80000000,%edx # FL_TUPLE_BIT
    jz setcarry
    jmp clrcarry

    # rsi=addr, r12=G
trynext: 
    movq %rsi,(%r12)       # goal->addr
    addq $8,%rsp
    jmp *%rsi

    .global fl_isval
    # rdx=val, rsi=addr
fl_isval: 
    cmpq %rdx,%rax
    jne trynext
    ret

    .global fl_isfloat
    # rax=val, rdx=float, rsi=addr 
fl_isfloat: 
    testq $3,%rax
    jz fl_isfloat_l1
    jmp trynext
fl_isfloat_l1: 
    cmpq %rdx,8(%rax)
    jne trynext
    ret

    .global fl_islist
    # rax=val, rsi=addr -> rax(car)
    # push cdr on stack
fl_islist: 
    call fl_islist_t
    jc trynext
    movq 16(%rax),%rdx
    xchgq %rdx,(%rsp)
    movq 8(%rax),%rax
    jmp *%rdx

    .global fl_isfunctor
    # rax=val, rsi=addr, rcx=arity, rdx=atom -> rax
fl_isfunctor: 
    testq $3,%rax
    jnz trynext
    movl (%rax), %r8d
    testl $0x80000000, %r8d
    jz trynext
    shrl $24, %r8d
    andl $0x7f, %r8d
    cmpl %ecx, %r8d
    jne trynext
    movq 8(%rax),%rcx
    cmpq %rdx,%rcx
    jne trynext
    movq 16(%rax),%rax
    ret

    # rbx=tcb, r10=car, rax=cdr -> rax
    # must preserve parameter registers
cons: 
    movq 6 * 8(%rbx), %r9
    testq %r9, %r9
    jnz cons_l1
    movq $10,%rdx           # FL_NO_CELLS
    jmp error
cons_l1: 
    pushq %rax
    movq 16(%r9), %rax
    movq %rax,6*8(%rbx)     # tcb->freelist
    popq %rax
    call fl_addref
    movq %rax, 16(%r9)
    movq %r10, %rax
    call fl_addref
    movq %rax, 8(%r9)
    incq 11*8(%rbx)            # tcb->used
    movq %r9, %rax
    ret

    .global fl_mkfloat
    # rdx=float64
fl_mkfloat: 
    pushq %rdx
    movq $1,%rax
    movq $1,%r10
    call cons
    popq %r9
    movl $0x02000000,(%rax)         # FL_FLOAT_TAG
    movq %r9, 8(%rax)
    ret

    .global fl_mklist
    # rax=cdr, [rsp+8]=car -> rax
fl_mklist: 
    movq 8(%rsp), %r10
    call cons
    movq $0x01000000, %r9
    movq %r9, (%rax)
    popq %r10
    addq $8,%rsp
    jmp *%r10

    .global fl_mktuple
    # rax=args, r10=atom, r9=arity -> rax
fl_mktuple: 
    pushq %r9
    call cons
    popq %r9
    shl $24, %r9
    orl $0x80000000, %r9d
    movl %r9d, (%rax)
    ret

    .global fl_mkvar
    # -> rax
fl_mkvar: 
    movq $1, %r10
    movq fl_nil,%rax
    call cons
    movq %rax,8(%rax)
    movl $0x04000000,(%rax)       # FL_VAR_TAG
    ret

    # r13=goal, rbx=tcb
enqueue: 
    movq 13*8(%rbx),%rax        # tcb->qend
    movq %r13, (%rax)
    addq $8,%rax
    movq 12*8(%rbx),%rcx        # tcb->qstart
    cmpq %rcx,%rax
    jne enqueue_l1
    movq $7,%rdx                # FL_QUEUE_FULL
    jmp error
enqueue_l1: 
    movq 15*8(%rbx),%rcx        # tcb->max_goals
    shlq $3,%rcx
    addq 14*8(%rbx),%rcx        # tcb->queue
    cmpq %rcx,%rax
    jb enqueue_l2
    movq 14*8(%rbx),%rax        # tcb->queue
enqueue_l2: 
    movq %rax,13*8(%rbx)        # tcb->qend
    incq 10*8(%rbx)             # tcb->active
    ret

    .global fl_opengoal
    # -> r13
fl_opengoal: 
    movq 16 * 8(%rbx), %r13
    testq %r13, %r13
    jne fl_opengoal_l1
    movq $26,%rdx               # FL_TOO_MANY_GOALS
    jmp error
fl_opengoal_l1: 
    movq (%r13), %r9
    movq %r9,  16 * 8(%rbx)
    incq 20*8(%rbx)            # tcb->goals
    jmp enqueue

    .global fl_dropgoal
    # rbx=tcb
fl_dropgoal: 
    movq (%rbx), %r9
    movq 16 * 8(%rbx), %r10
    movq %r10, (%r9)
    movq %r9, 16 * 8(%rbx)
    decq 20*8(%rbx)            # tcb->goals
    ret

    .global fl_addref
    # rax=val
fl_addref: 
    testq $3,%rax
    jz fl_addref_l1
    ret
fl_addref_l1: 
    pushq %rax
    movl (%rax),%eax
    andl $0xffffff,%eax    # FL_COUNT_MASK
    cmpl $0xffffff,%eax    # FL_COUNT_MASK
    je fl_addref_l2
    popq %rax
    incl (%rax)
    ret
fl_addref_l2: 
    popq %rax
    ret

    .global fl_puttailarg
    # r13=N, rax=list
fl_puttailarg: 
    movq 16(%rax),%rdx
    testq $3,%rdx
    jz fl_puttailarg_l1
    movq 8(%rax),%rax
fl_puttailarg_l1: 
    call fl_addref
    movq %rax, 4 * 8(%r13)
    ret

    .global fl_resolve
    # r8=cache, rsi=arity, rdx=module, rcx=name
    # r13=N
fl_resolve: 
    movq (%r8), %r9
    testq %r9, %r9
    je fl_resolve_l1
    movq %r9, (%r13)
    ret
fl_resolve_l1: 
    movq %rbx,%rdi
    movq %r13, %r9
    jmp fl_resolve_pdef

    .global fl_replace
    # rax=val, rdx=argaddr
fl_replace: 
    cmpq (%rdx),%rax
    jne fl_replace_l1
    ret
fl_replace_l1: 
    movq (%rdx), %r9
    call fl_addref
    movq %rax,(%rdx)
    jmp fl_unref

    .global fl_tail_goal
    # rdx=addr, rbx=tcb -> r13(N)
fl_tail_goal: 
    movq 16(%rbx),%rcx  # tcb->tailcalls
    movq 24(%rbx),%rsi  # tcb->timeslice
    cmpq %rsi,%rcx
    jl fl_invoke_goal
    pushq %rdx
    call fl_opengoal
    popq %rdx
    movq %rdx,(%r13)       # goal->addr
    ret

    .global fl_invoke_goal
    # rdx=addr, rbx=tcb -> r13(N)
fl_invoke_goal: 
    movq %r12, %r13
    movq %rdx, (%r13)       # goal->addr
    ret

    .global fl_tail_jump
    # r13=r12=N, rbx=tcb, rsp=frame
fl_tail_jump: 
    movq 16(%rbx),%rcx  # tcb->tailcalls
    movq 24(%rbx),%rsi  # tcb->timeslice
    cmpq %rsi,%rcx
    jl fl_tail_jump_l1
    xorq %rax,%rax
    movq %rax,16(%rbx)  # tcb->tailcalls
    ret
fl_tail_jump_l1: 
    incq %rcx
    movq %rcx,16(%rbx)  # tcb->tailcalls
    # fall through ...

    .global fl_invoke_jump
    # r13=r12=N, rbx=tcb, rsp=frame
fl_invoke_jump: 
    movq %rbp,%rsp
    movq %r13, %r12
    movq %r13, (%rbx)
    movq (%r13), %rax       # goal->addr
    jmp *%rax

    .global fl_test_integer
    # rsi=val -> rax (bool)
fl_test_integer: 
    testq $1,%rsi
    jnz true
    jmp false

    .global fl_test_number
    # rsi=val -> rax (bool)
fl_test_number: 
    testq $1,%rsi
    jnz true
    movl $0x02000000, %r9d
    jmp fl_test_tag

    .global fl_test_atom
    # rsi=val -> rax (bool)
fl_test_atom: 
    movq %rsi,%rax
    andq $3,%rax
    cmpq $2,%rax
    jne false
    jmp true

    .global fl_test_tag
    # rsi=val, r9d=tag -> rax (bool)
fl_test_tag: 
    testq $3,%rsi
    jnz false
    movl (%rsi),%eax
    andl $0xff000000,%eax # FL_TAG_MASK
    cmpl %r9d, %eax
    jne false
    jmp true

    .global fl_test_var
    # rsi=val -> rax (bool)
fl_test_var: 
    testq $3,%rsi
    jnz false
    movl (%rsi),%eax
    andl $0xff000000,%eax # FL_TAG_MASK
    cmpl $0x07000000,%eax # FL_REF_TAG
    je true
    cmpl $0x04000000,%eax # FL_VAR_TAG
    jne false
    movq 8(%rsi),%rax
    cmpq %rsi,%rax
    jne false
    jmp true

    .global fl_test_tuple
    # rsi=val -> rax (bool)
fl_test_tuple: 
    testq $3,%rsi
    jnz false
    movl (%rsi),%eax
    testl $0x80000000,%eax # FL_TUPLE_BIT
    jz false
    jmp true

    .global fl_enter
    # rdi=tcb
fl_enter: 
    movq 5*8(%rdi),%rsp     # tcb->stackbase
    movq %rsp,%rbp
    movq (%rdi), %r12       # tcb->goal
    movq (%r12), %rax       # goal->addr
    movq %rdi,%rbx
    jmp *%rax

    .global fl_atomic
    # rsi=val -> rax (bool)
fl_atomic: 
    testq $3,%rsi       # FL_BITS_MASK
    jnz true
    movl (%rsi),%eax
    andl $0xff000000,%eax   # FL_TAG_MASK
    cmpl $0x02000000,%eax   # FL_FLOAT_TAG
    je true
    jmp false

    .global fl_entry
    # rbx=tcb, rsi=arity
fl_entry: 
    movl 19*8(%rbx),%eax        # tcb->logging
    testl $1,%eax
    jnz fl_entry_l1
    ret
fl_entry_l1: 
    movq %rbx,%rdi
    jmp fl_log_entry
    ret

    .global fl_unify
    # rsi=val, rdx=val -> rax (bool)
    # must preserve P0-P3
fl_unify: 
    pushq %rsi
    pushq %rdx
    pushq %rcx
    pushq %r8
    movq %rbx,%rdi
    subq $8,%rsp            # (un)align
    call fl_unify_rec
    addq $8,%rsp
    popq %r8
    popq %rcx
    popq %rdx
    popq %rsi
    ret
