% FLENG compiler - toplevel

:- dynamic([pdef/2, r_atom/2, r_data/1, module_name/1, module_init/1]).
:- dynamic([no_warnings/0, debug_info/0, entry_point/2, uses/1]).
:- dynamic([drop_sequence/3, literal/2, symbol_loc/2]).
:- dynamic([pdef_name/2, module_info/3]).
:- initialization(main).

:- if(current_prolog_flag(dialect, gprolog)).
% GNU
:- include('version.pl').
:- include('compiler.pl').
:- include('index.pl').
:- include('lfa.pl').
:- include('asm.pl').

main :-
    argument_list(Args),
    main(Args).
:- endif.

:- if(current_prolog_flag(dialect, swi)).
% SWI
:- include(fleng('version.pl')).
:- include(fleng('compiler.pl')).
:- include(fleng('index.pl')).
:- include(fleng('lfa.pl')).
:- include(fleng('asm.pl')).

format_to_atom(Out, Fstr, Args) :-
    format_to_codes(Fstr, Args, S),
    atom_codes(Out, S).

g_assign(Var, Val) :- nb_setval(Var, Val).
g_read(Var, Val) :- nb_getval(Var, Val).
:- endif.

:- if(current_prolog_flag(dialect, yap)).
% YAP
:- use_module(library(lists)).
:- use_module(library(charsio)).
:- include('version.pl').
:- include('compiler.pl').
:- include('index.pl').
:- include('lfa.pl').
:- include('asm.pl').

start :-
    set_prolog_flag(syntax_errors, error),
    set_prolog_flag(unknown, error),
    unix(argv(Args)),
    catch(main(Args), E, quit(E)).

subtract([], _, []).
subtract([Element|Residue], Set, Difference) :-
	memberchk(Element, Set), !,
	subtract(Residue, Set, Difference).
subtract([Element|Residue], Set, [Element|Difference]) :-
	subtract(Residue, Set, Difference).

numbervars(Term) :- numbervars(Term, 0, _).

format_to_atom(Out, Fstr, Args) :-
    format_to_chars(Fstr, Args, S),
    atom_codes(Out, S).

save_to_file(Name) :- save(Name), halt(0).

quit(E) :-
    print_message(error, E),
    halt(1).

g_assign(Var, Val) :- nb_setval(Var, Val).
g_read(Var, Val) :- nb_getval(Var, Val).
:- endif.
